<?php
	if (!isset($argv[1])) {
		die("Usage: {$argv[-1]} #instance");
	}
	
	$numero = (int)$argv[1];
	$instancia = json_decode(@file_get_contents("instance_data_{$numero}.json.txt"), true);
	$consumo = json_decode(@file_get_contents("power_data_{$numero}.json.txt"), true);
	
	if (!isset($instancia, $consumo)) {
		die("{$argv[0]} {$argv[1]}: unable to read the input files");
	}
	
	// reconstruction of the input
		
	$corridas = [ ];
	
	foreach ($instancia['Trains'] as $trenes) {
		foreach ($trenes['Legs'] as $corrida) {
			$corridas[$corrida['LegID']] = $corrida + [ 'TrainID' => $trenes['TrainID'] ];
		}
	}
	
	foreach ($consumo['Powerprofiles'] as $onda) {
		$corridas[$onda['LegID']]['Powerprofiles'] = $onda['Powerprofile'];
	}

	unset($instancia);
	unset($consumo);

	// LP file
	ob_start( ); {
		// objective function definition
		
		echo "minimize\n";
		echo "\tpromedio_intervalo\n";
		
		// constraints
		
		echo "subject to\n";
		
		function mapea($corridas, $atributo)
		{
			$arr = [ ];
			
			foreach ($corridas as $corrida) {
				$arr[$corrida[$atributo]][] = $corrida;
			}
			
			return $arr;
		}
		
		function compara_3vias($a, $b)
		{
			return 2 * ($a > $b) - ($a != $b);
		}
		
		function compara_llegadas($corrida1, $corrida2)
		{
			return compara_3vias($corrida1['CurrentDepartureTime'] + $corrida1['TravelTime'], $corrida2['CurrentDepartureTime'] + $corrida2['TravelTime']);
		}
		
		function compara_salidas($corrida1, $corrida2)
		{
			return compara_3vias($corrida1['CurrentDepartureTime'], $corrida2['CurrentDepartureTime']);
		}
		
		// constraints - headways
		
		echo "\\\theadways\n";
		
		foreach (mapea($corridas, 'TrackID') as $corridas_via) {
			usort($corridas_via, 'compara_salidas');
		
			for ($i = 1; $i < count($corridas_via); ++$i) {
				$corrida = $corridas_via[$i];
				$anterior = $corridas_via[$i - 1];
				
				echo "\tsalida_{$corrida['LegID']} - salida_{$anterior['LegID']} >= {$anterior['MinimumHeadwayTime']}\n";
			}
		}
		
		// constraints - stops at stations
		
		echo "\\\tstops at stations\n";
		
		foreach (mapea($corridas, 'TrainID') as $corridas_tren) {
			usort($corridas_tren, 'compara_salidas');

			for ($i = 1; $i < count($corridas_tren); ++$i) {
				$corrida = $corridas_tren[$i];
				$anterior = $corridas_tren[$i - 1];
				$viaje_con_parada = $anterior['MinimumStoppingTime'] + $anterior['TravelTime'];

				echo "\tsalida_{$corrida['LegID']} - salida_{$anterior['LegID']} >= {$viaje_con_parada}\n";
			}
		}
		
		// constraints - connections
		
		echo "\\\tconnections\n";
		
		$llegadas = mapea($corridas, 'EndStationID');
		$salidas = mapea($corridas, 'StartStationID');
		$estaciones = array_unique(array_merge(array_keys($llegadas), array_keys($salidas)));
		
		foreach ($estaciones as $estacion) {
			if (!isset($llegadas[$estacion])) {
				$llegadas[$estacion] = [ ];
			}
			
			if (!isset($salidas[$estacion])) {
				$salidas[$estacion] = [ ];
			}
			
			usort($llegadas[$estacion], 'compara_llegadas');
			usort($salidas[$estacion], 'compara_salidas');
			
			foreach ($salidas[$estacion] as $salida) {
				foreach ($llegadas[$estacion] as $llegada) {
					$diferencia = $salida['CurrentDepartureTime'] - ($llegada['CurrentDepartureTime'] + $llegada['TravelTime']);
					
					if ($diferencia < 5) {
						break;
					}
					
					if ($diferencia > 15) {
						continue;
					}
					
					if ($salida['TrainID'] != $llegada['TrainID']) {
						$inferior = $llegada['TravelTime'] + 5;
						$superior = $llegada['TravelTime'] + 15;
						
						echo "\tsalida_{$salida['LegID']} - salida_{$llegada['LegID']} >= {$inferior}\n";
						echo "\tsalida_{$salida['LegID']} - salida_{$llegada['LegID']} <= {$superior}\n";
					}
				}
			}
		}

		// linearization constraints - preprocessing

		$intervalos = [ ];
		
		function accede($onda, $offset)
		{
			return ($offset < 0 || $offset >= count($onda) ? 0 : $onda[$offset]);
		}
		
		foreach ($corridas as $corrida) {
			$pri = $corrida['EarliestDepartureTime'];
			$ult = $corrida['LatestDepartureTime'];

			for ($i = $pri; $i <= $ult; ++$i) {
				$bloque_ini = (int)($i / 15);
				$bloque_ult = (int)(($i + $corrida['TravelTime']) / 15);

				for ($j = $bloque_ini; $j <= $bloque_ult; ++$j) {
					$intervalos[$j][$corrida['LegID']][] = $i;
				}
			}
		}

		// linearization constraints - power consumption
		
		echo "\\\tlinearization constraints - power consumption\n";
		
		foreach ($intervalos as $bloque => $datos) {
			$segundo_ini = 15 * 60 * $bloque;
			$segundo_fin = 15 * 60 * ($bloque + 1);
			$sumandos = [ 'promedio_intervalo' ];
			
			for ($i = $segundo_ini; $i <= $segundo_fin; ++$i) {
				$contribuciones = [ ];
				
				foreach ($datos as $indice_corrida => $salidas) {
					$corrida = $corridas[$indice_corrida];
					
					foreach ($salidas as $salida) {
						$contribucion = accede($corrida['Powerprofiles'], $i - 60 * $salida) / (1 + (int)($i % 900 == 0));
						
						if ($contribucion != 0) {
							$contribuciones[] = "{$contribucion} salida_{$corrida['LegID']}_{$salida}";
						}
					}
				}
				
				if (empty($contribuciones)) {
					echo "\tsegundo_{$i} = 0\n";
				}
				else {
					echo "\t", implode(' + ', $contribuciones), " - segundo_{$i} <= 0\n";
				}
				
				$sumandos[] = "segundo_{$i}";
			}
			
			echo "\t", implode(' - ', $sumandos), " >= 0\n";
		}
		
		// linearization constraints - integer to binary variables
		
		echo "\\\tlinearization constraints - integer to binary variables\n";
		
		foreach ($corridas as $corrida) {
			$suma1 = [ ];
			$sumap = [ ];
			
			$pri = $corrida['EarliestDepartureTime'];
			$ult = $corrida['LatestDepartureTime'];

			for ($i = $pri; $i <= $ult; ++$i) {
				$suma1[] = "salida_{$corrida['LegID']}_{$i}";
				$sumap[] = "{$i} salida_{$corrida['LegID']}_{$i}";
			}
			
			echo "\t", implode(' + ', $suma1), " = 1\n";
			echo "\t", implode(' + ', $sumap), " - salida_{$corrida['LegID']} = 0\n";
		}
		
		// bounds
		
		echo "bounds\n";
		
		foreach ($corridas as $corrida) {
			echo "\t{$corrida['EarliestDepartureTime']} <= salida_{$corrida['LegID']} <= {$corrida['LatestDepartureTime']}\n";
		}
		
		$ultimo_segundo = (max(array_keys($intervalos)) + 1) * 15 * 60;
		
		for ($i = 0; $i <= $ultimo_segundo; ++$i) {
			echo "\tsegundo_{$i} >= 0\n";
		}
		
		// integer variables
		
		echo "general\n";
		
		foreach ($corridas as $corrida) {
			echo "\tsalida_{$corrida['LegID']}\n";
		}
		
		// binary variables
		
		echo "binary\n";
		
		foreach ($corridas as $corrida) {
			$pri = $corrida['EarliestDepartureTime'];
			$ult = $corrida['LatestDepartureTime'];
				
			for ($i = $pri; $i <= $ult; ++$i) {
				echo "\tsalida_{$corrida['LegID']}_{$i}\n";
			}
		}
		
		// end
		
		echo "end\n";
	} file_put_contents("model{$numero}.lp", ob_get_clean( ));
	
	// MST file
	ob_start( ); {
		foreach ($corridas as $corrida) {
			echo "salida_{$corrida['LegID']} {$corrida['CurrentDepartureTime']}\n";
		}
	} file_put_contents("model{$numero}.mst", ob_get_clean( ));
?>
